/*
 *    _           _             _   _
 *   (_) __ _  __| | __ _ _ __ | |_(_)_   _____
 *   | |/ _` |/ _` |/ _` | '_ \| __| \ \ / / _ \
 *   | | (_| | (_| | (_| | |_) | |_| |\ V /  __/
 *  _/ |\__,_|\__,_|\__,_| .__/ \__|_| \_/ \___|
 * |__/                  |_|
 *
 * This file is part of the Maverick Synergy Hotfixes Java SSH API
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited.
 *
 * Copyright (C) 2002-2021 JADAPTIVE Limited - All Rights Reserved
 *
 * Use of this software may also be covered by third-party licenses depending on the choices you make about what features to use.
 *
 * Please visit the link below to see additional third-party licenses and copyrights
 *
 * https://www.jadaptive.com/app/manpage/en/article/1565029/What-third-party-dependencies-does-the-Maverick-Synergy-API-have
 */

package com.sshtools.common.files.vfs;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;

import com.sshtools.common.files.AbstractFile;
import com.sshtools.common.files.AbstractFileFactory;
import com.sshtools.common.logger.Log;
import com.sshtools.common.permissions.PermissionDeniedException;
import com.sshtools.common.util.FileUtils;

public class VirtualMountManager {

	private VirtualMount defaultMount;
	private List<VirtualMount> mounts = new ArrayList<VirtualMount>();
	private VirtualFileFactory fileFactory;
	private ThreadLocal<VirtualMount> testingMount = new ThreadLocal<>(); 
	
	public VirtualMountManager(VirtualFileFactory fileFactory) throws IOException,
			PermissionDeniedException {
		this(fileFactory, null);
	}

	public VirtualMountManager(VirtualFileFactory fileFactory, VirtualMountTemplate homeMount,
			VirtualMountTemplate... additionalMounts) throws IOException,
			PermissionDeniedException {

		this.fileFactory = fileFactory;

		if (homeMount != null) {
			defaultMount = new VirtualMount(homeMount, 
					fileFactory, homeMount.getActualFileFactory(), 
					true, false, homeMount.isCreateMountFolder(),
					homeMount.lastModified());
			if(defaultMount.isCreateMountFolder()) {
				defaultMount.getActualFileFactory().getFile(defaultMount.getRoot()).createFolder();
			}
			mounts.add(defaultMount);
		}

		// Add any remaining templates
		for (VirtualMountTemplate m : additionalMounts) {
			VirtualMount vm = createMount(m,
					m.getActualFileFactory(),
					m.isCreateMountFolder(),
					m.lastModified());
			

			if(vm.isCreateMountFolder()) {
				vm.getActualFileFactory().getFile(vm.getRoot()).createFolder();
			}
			mounts.add(vm);
		}

		sort(mounts);
	}
	
	public void mount(VirtualMountTemplate template) throws IOException, PermissionDeniedException {
		mount(template, false);
	}
	
	public void mount(VirtualMountTemplate template, boolean unmount) throws IOException, PermissionDeniedException {
		mount(createMount(template, 
				template.getActualFileFactory(),
				template.isCreateMountFolder(),
				template.lastModified()), unmount);
	}

	public void test(VirtualMountTemplate template) throws IOException, PermissionDeniedException {
		
		test(createMount(template, 
				template.getActualFileFactory(),
				template.isCreateMountFolder(),
				template.lastModified()));
		
	}
	
	private void test(VirtualMount mount) throws IOException, PermissionDeniedException {
		
		Log.info("Testing " + mount.getMount() + " on " + mount.getRoot());

		try {
			testingMount.set(mount);
			AbstractFile f = fileFactory.getFile(mount.getMount());
			if(mount.isCreateMountFolder()) {
				f.createFolder();
			}
		} catch (Exception ex) {
			Log.error("Cannot mount " + mount.getMount() + " " + mount.getRoot(), ex);
		} finally {
			testingMount.remove();
		}
	}
	
	private void mount(VirtualMount mount, boolean unmount) throws IOException,
			PermissionDeniedException {
		
		if(unmount && isMounted(mount.getMount())) {
			unmount(mount);
		}

		if(isMounted(mount.getMount())) {
			throw new IOException(mount.getMount() + " already mounted on " + getMount(mount.getMount()).getRoot());
		} 
		
		Log.info("Mounting " + mount.getMount() + " on " + mount.getRoot());

		test(mount);
		
		// Add the mount
		mounts.add(mount);
		sort(mounts);

		Log.info("Mounted " + mount.getMount() + " on " + mount.getRoot());

	}

	private void sort(List<VirtualMount> mounts) {
		Collections.sort(mounts, new Comparator<VirtualMount>() {

			@Override
			public int compare(VirtualMount o1, VirtualMount o2) {
				if(o1.isParentOf(o2)) {
					return 1;
				} else if(o1.isChildOf(o2)) {
					return -1;
				} else {
					return o2.getMount().compareTo(o1.getMount());
				}
			}
			
		});
		
		if(Log.isDebugEnabled()) {
			Log.debug("Sorting mounts by path and with child relationship preferred");
			for(VirtualMount m : mounts) {
				Log.debug("Mount {} on {}", m.getMount(), m.getRoot());
			}
		}
	}

	public void unmount(VirtualMount mount) throws IOException {
		Log.info("Unmounting " + mount.getMount() + " from " + mount.getRoot());
		VirtualMount mounted = null;
		for(VirtualMount m : mounts) {
			if(FileUtils.checkEndsWithSlash(m.getMount())
					.equals(FileUtils.checkEndsWithSlash(mount.getMount()))) {
				mounted = m;
			}
		}
		if(Objects.isNull(mounted)) {
			throw new IOException(String.format("Could not find mount %s", mount.getMount()));
		}
		mounts.remove(mounted);
		sort(mounts);
		
	}

	public VirtualMount getDefaultMount() {
		return defaultMount;
	}

	public VirtualMount[] getMounts() {
		List<VirtualMount> tmp = new ArrayList<>();
		VirtualMount testMount = testingMount.get();
		if(testMount!=null) {
			tmp.add(testingMount.get());
		}
		tmp.addAll(mounts);
		sort(tmp);
		return tmp.toArray(new VirtualMount[0]);
	}

	public boolean isMounted(String path) {
		if (path.equals("")) {
			return true;
		}

		for (VirtualMount mount : mounts) {
			if (FileUtils.addTrailingSlash(mount.getMount()).equals(
					FileUtils.addTrailingSlash(path))) {
				return true;
			}
		}
		return false;
	}

	private VirtualMount createMount(VirtualMountTemplate template,
			AbstractFileFactory<?> actualFileFactory, boolean createMoundFolder, long lastModified) throws IOException,
			PermissionDeniedException {
		return new VirtualMount(template, fileFactory, actualFileFactory, createMoundFolder, lastModified);
	}

	public VirtualMount getMount(String path) throws IOException {

		path = path.replace('\\', '/').trim();

		if (path.equals("") || path.equals(".") || path.startsWith("./")) {
			return defaultMount;
		}

		for (VirtualMount mount : getMounts()) {
			String mountPath = FileUtils.checkEndsWithSlash(mount.getMount());
			path = FileUtils.checkEndsWithSlash(path);
			if (path.startsWith(mountPath)) {
				return mount;
			}
		}
		throw new FileNotFoundException("No mount for " + path);
	}

	public VirtualMount[] getMounts(String path) {
		if (path.equals("")) {
			return new VirtualMount[] { defaultMount };
		}

		path = FileUtils.addTrailingSlash(path);

		List<VirtualMount> matched = new ArrayList<VirtualMount>();
		for (VirtualMount m : getMounts()) {
			String mountPath = FileUtils.addTrailingSlash(m.getMount());
			if (path.startsWith(mountPath) || mountPath.startsWith(path)) {
				matched.add(m);
			}
		}		
		
		return matched.toArray(new VirtualMount[0]);
	}

	public VirtualFileFactory getVirtualFileFactory() {
		return fileFactory;
	}

}